package cc.smtweb.system.bpm.web.engine.flow;

import cc.smtweb.framework.core.common.SwMap;
import cc.smtweb.framework.core.db.DbEngine;
import cc.smtweb.framework.core.db.EntityDao;
import cc.smtweb.framework.core.db.EntityHelper;
import cc.smtweb.framework.core.exception.BizException;
import cc.smtweb.framework.core.exception.SwException;
import cc.smtweb.framework.core.session.UserSession;
import cc.smtweb.framework.core.util.CommUtil;
import cc.smtweb.framework.core.util.DateUtil;
import cc.smtweb.framework.core.util.NumberUtil;
import cc.smtweb.system.bpm.web.design.flow.FlowConst;
import cc.smtweb.system.bpm.web.design.flow.ModelProc;
import cc.smtweb.system.bpm.web.design.flow.ModelProcHelper;
import cc.smtweb.system.bpm.web.design.flow.define.Activity;
import cc.smtweb.system.bpm.web.engine.flow.entity.*;
import cc.smtweb.system.bpm.web.sys.user.dept.Dept;
import cc.smtweb.system.bpm.web.sys.user.dept.DeptCache;
import cc.smtweb.system.bpm.web.sys.user.party.Party;
import cc.smtweb.system.bpm.web.sys.user.party.PartyCache;
import cc.smtweb.system.bpm.web.sys.user.user.UserCache;
import lombok.Data;
import org.apache.commons.lang3.StringUtils;

import java.util.*;

/**
 * Created by Akmm at 2020/3/4 11:35
 * 单据流程实例组件
 */
@Data
public class FlowInstance {
    //会话信息
    private UserSession us;
    //流程定义
    private ModelProc procDef;
    //流程实例
    private ProcInst procInst;

    //活动任务
    private Task actInst;

    //操作类型 新增/编辑/查看
    private int opt_mode = FlowConst.OperatorType.VIEW;

    //流程执行变量
//    public Map<String, String> variables = new HashMap<>();

    //当前任务
    private List<Map<String, String>> tasks = new ArrayList<>();
    //数据提供者
    private FlowProvider provider = new FlowProvider();

    public FlowInstance(UserSession us) {
        this.us = us;
    }

    //能否编辑
    public boolean canEdit() {
        //新增状态可以编辑
        if (opt_mode == FlowConst.OperatorType.NEW || procInst.getStatu() == FlowConst.InstanceStatu.BEGIN.value)
            return true;
        return false;
    }


    /**
     * 流程条件
     *
     * @param expr
     * @return
     */
    public boolean execTrans(String expr) {
        if (StringUtils.isBlank(expr)) return true;
        Map<String, Object> mapVar = new HashMap<>();
        /*mapVar.putAll(billEntity.getData());
        mapVar.put("party_code", PartyEntityBuffer.getInstance().getCodeById(billEntity.getPartyId()));
        mapVar.put("party_name", PartyEntityBuffer.getInstance().getNameById(billEntity.getPartyId()));
        mapVar.put("dept_code", DepartmentEntityBuffer.getInstance().getCodeById(billEntity.getDeptId()));
        mapVar.put("dept_name", DepartmentEntityBuffer.getInstance().getNameById(billEntity.getDeptId()));
        mapVar.put("bill_user_code", UserBaseEntityBuffer.getInstance().getCodeById(billEntity.getBillUser()));
        mapVar.put("bill_user_name", UserBaseEntityBuffer.getInstance().getNameById(billEntity.getBillUser()));
        localBuildExecTrans(mapVar);*/
        Object o = NumberUtil.calcExprMapObject(expr, mapVar);
        return o != null && (Boolean) o;
    }

    /**
     * 新建流程实例
     *
     * @param bill_type
     * @throws Exception
     */
    public void create(int bill_type) {
        Party loginParty = PartyCache.getInstance().get(us.getPartyId());
        if (EntityHelper.isEmpty(loginParty)) {
            throw new BizException("当前登录人员管理单位信息为空，请检查后保存！");
        }

        Dept loginDept = DeptCache.getInstance().get(us.getDeptId());
        if (EntityHelper.isEmpty(loginDept)) {
            loginDept = new Dept();
        }

        procDef = ModelProcHelper.getBillProc(bill_type, us.getUserId()); //根据单据分类获取流程定义
        if (EntityHelper.isEmpty(procDef)) {
//            throw new BizException("没有找到对应的流程定义，请到【系统设置】-【单据流程分配】为当前单据类型【" + BillTypeCache.getInstance().getName(bill_type) + "】分配工作流程！");
        }

        Activity start = ModelProcHelper.getFirstActivity(procDef.getEntityId());
        if (start == null) {
            throw new BizException("对应的流程没有定义有效的活动步骤！");
        }

        opt_mode = FlowConst.OperatorType.NEW;

        procInst = new ProcInst();
        procInst.init();
        procInst.setEntityId(DbEngine.getInstance().nextId());
        procInst.setPartyId(loginParty.getEntityId());
        procInst.setDeptId(loginDept.getEntityId());
        procInst.setBillType(bill_type);
        procInst.setMakeDate(DateUtil.nowDateTimeLong());
        procInst.setPrcId(procDef.getEntityId());
        procInst.setStartTime(DateUtil.nowDateTimeLong());
        procInst.setUserId(us.getUserId());
        procInst.setStatu(FlowConst.InstanceStatu.BEGIN.value);

        actInst = new Task();
        actInst.init();
        actInst.setEntityId(DbEngine.getInstance().nextId());
        actInst.setId(procInst.getEntityId());
        actInst.setPrcId(procDef.getEntityId());
        actInst.setActId(start.getId());
        actInst.setHandler(us.getUserId());
        actInst.setStartTime(DateUtil.nowDateTimeLong());
        actInst.setStatu(FlowConst.ActivityStatu.HANDLE.value);
        actInst.setMake(true);
        actInst.setSign(false);
        actInst.setReject(false);
        actInst.setRetake(false);

        procInst.setTaskId(actInst.getEntityId());
    }

    /**
     * 办理
     *
     * @throws Exception
     */
    public void handler() {
        long user_id = us.getUserId();
        //校验权限，当前人员有权限干不，没权限直接抛异常
//        checkPermission();
        if (actInst.getStatu() != FlowConst.ActivityStatu.WAIT.value) {
            throw new BizException("当前任务非待办状态，不能办理失败！");
        }

        actInst.setHandler(user_id);
        actInst.setStartTime(DateUtil.nowDateTimeLong());
        actInst.setStatu(FlowConst.ActivityStatu.HANDLE.value);

        procInst.setTaskId(actInst.getEntityId());
        procInst.setStatu(FlowConst.InstanceStatu.RUNING.value);

        DbEngine.getInstance().findDao(Task.ENTITY_NAME).updateEntity(actInst);
        DbEngine.getInstance().findDao(ProcInst.ENTITY_NAME).updateEntity(procInst);

        buildBillLog(FlowConst.Button.HANDLE.value, actInst, null);
//        comp.transCallback(FlowConst.FlowOptType.RETAKE.value, act_inst, null);
    }

    //从页面参数加载流程实例
    public void readFromPage(SwMap map) {
        procInst.readFromMap(map);
        actInst.setPriId(procInst.getId());
        actInst.setEntityId(procInst.getTaskId());
    }

    /**
     * 从数据库加载已有流程实例
     *
     * @param bill_id
     * @throws Exception
     */
    public void load(long bill_id, long task_id) {
        procInst = provider.findProcInst(bill_id);
        procDef = ModelProcHelper.getFromCache(procInst.getPrcId());
        task_id = getTaskId(task_id);
        if (task_id > 0L) {
            actInst = provider.findTask(task_id);
        }
    }

    /**
     * 当前任务处于待办时：
     * 上一步办理人进来，看上一个步骤，否则看当前步骤
     *
     * @param task_id
     * @return
     */
    private long getTaskId(long task_id) {
        if (task_id > 0L) return task_id;
        Task curTask = provider.findTask(task_id);
        if (EntityHelper.isEmpty(curTask) || curTask.getStatu() != FlowConst.ActivityStatu.WAIT.value) return task_id;
        Task preAct = provider.findBeforeTask(task_id);
        if (EntityHelper.isEmpty(preAct)) return task_id;
        if (preAct.getHandler() == us.getUserId()) {
            return preAct.getId();
        }
        return task_id;
    }

    /**
     * 矫正默认任务及获取可选任务列表，供界面选择
     *
     * @return
     * @throws Exception
     */
    public List<Task> adjustTasks() {
        List<Task> listTask = new ArrayList<>();
        if (opt_mode == FlowConst.OperatorType.NEW || procInst.getStatu() == FlowConst.InstanceStatu.DISUSE.value) {
            listTask.add(actInst);
            return listTask;
        }
        //获取当前单据活动任务列表及当前任务
//        act_inst = null;
        List<Task> list = provider.findCurTask(procInst.getId());
        listTask.addAll(list);
        //当前登录用户
        String sqlIn = "";
        for (Task act : list) {
            //如果办理人是当前用户，则置此任务为当前任务
//            if (act_inst == null) {
            //我办理中的任务：非待办，且是我的任务
            if (act.getStatu() != FlowConst.ActivityStatu.WAIT.value) {
                if (us.getUserId() == act.getHandler()) {
                    actInst = act;
                }
            } else if (FlowHelper.isMyWaitedTask(act.getId(), us.getUserId())) {
                //我的待办任务
                actInst = act;
                act.setHandler(us.getUserId());
            } else {
                Activity ae = procDef.getProcInfo().findActivity(act.getActId());
                if (ae != null && ae.getType() == FlowConst.ActivityType.END.value) {
                    actInst = act;
                }
            }
//            }
            //如果当前任务为待办理，那么要找其前置任务，作为活动任务，方便取回
            if (act.getStatu() == FlowConst.ActivityStatu.WAIT.value) {
                sqlIn += "," + act.getEntityId();
            }
        }
        //找待办任务的前置任务，便于取回
        if (StringUtils.isNotEmpty(sqlIn)) {
            List<Task> list1 = provider.findBeforeTasks(sqlIn.substring(1));
            listTask.addAll(list1);
            if (actInst == null) {
                for (Task act : list1) {
                    //如果办理人是当前用户，则置此任务为当前任务
                    if (us.getUserId() == act.getHandler()) {
                        actInst = act;
                        break;
                    }
                }
            }
        }
        //没有我的处理任务，用流程当前任务来
        if (actInst == null) {
            actInst = provider.findTask(procInst.getTaskId());
        }

        return listTask;
    }

    /**
     * 设置单据描述及链接地址，待办中展示
     *
     * @param bill_info
     * @param bill_url
     */
    public void setBillInfo(String bill_code, String bill_info, String bill_url) {
        procInst.setBillInfo(bill_info);
        procInst.setBillCode(bill_code);
//        procInst.setBillUrl(bill_url); todo
    }

    //启动流程
    public void save() {
        if (opt_mode == FlowConst.OperatorType.NEW) {
            getProcInstDao().insertEntity(procInst);
            getTaskDao().insertEntity(actInst);
            buildBillLog(FlowConst.Button.ADD.value, null, null);
            opt_mode = FlowConst.OperatorType.EDIT;
        } else {
            getProcInstDao().updateEntity(procInst);
            if (actInst != null) {
                getTaskDao().updateEntity(actInst);
            }
        }
    }

    /**
     * 删除流程
     *
     * @throws Exception
     */
    public void delete() {
        getProcInstDao().deleteEntity(procInst);
        getTaskDao().deleteEntity("where bill_id=?", procInst.getId());
        getCandidateDao().deleteEntity("where bill_id=?", procInst.getId());
        getCareDao().deleteEntity("where bill_id=?", procInst.getId());
        getCommentDao().deleteEntity("where bill_id=?", procInst.getId());
        getSignDao().deleteEntity("where bill_id=?", procInst.getId());
        getTaskRelDao().deleteEntity("where bill_id=?", procInst.getId());
    }

    /**
     * 返回页面前的设置
     */
    public void resetPage() {
        List<Task> listTask = adjustTasks();
        //可选步骤
        tasks.clear();
        String next_text = procDef.getProcInfo().getActNameById(actInst.getActId());
        int index = 0;
        for (Task act : listTask) {
            Map<String, String> row = new HashMap<>();
            tasks.add(row);
            row.put("id", String.valueOf(act.getId()));
            if (act.getId() == actInst.getId() && listTask.size() > 1 && index != 0) {
                next_text = "待" + next_text;
            } else {
                next_text = procDef.getProcInfo().getActNameById(act.getActId());
            }
            row.put("text", next_text);
            index++;
        }
        resetButtons();
    }

    public void resetButtons() {
        Map<String, Integer> buttons = new HashMap<>();
        //按钮
        buttons.clear();
        long user_id = us.getUserId();
        /*
        //菜单权限
        Map<String, String> mapPermisson = MenuHelper.getPermission(context);
        //新增不要显示历史
        boolean isNew = opt_mode == FlowConst.OperatorType.NEW;
        boolean isAdmin = UserBaseEntityBuffer.getInstance().isAdmin(user_id);
        boolean canEdit = MenuHelper.hasPermisson(mapPermisson, FlowConst.MenuFuncRight.ADD.value) || MenuHelper.hasPermisson(mapPermisson, FlowConst.MenuFuncRight.UPD.value);
        //能否新增，菜单有授权+单据有流程分配
        boolean canAdd = canEdit && BillFlowWfHelper.getBillProc(procInst.getBillType(), user_id) != null;
        boolean canDel = MenuHelper.hasPermisson(mapPermisson, FlowConst.MenuFuncRight.DEL.value);
        boolean canDis = MenuHelper.hasPermisson(mapPermisson, FlowConst.MenuFuncRight.DIS.value) && canEdit && !isNew;
        //非制单人能否修改单据
        boolean othCanEdit = BillFlowWfHelper.getPropBool(procDef.getEntityId(), FlowConst.NULL_STR, FlowConst.ProcProperty.CAN_EDIT_OTHER);

        buttons.put(FlowConst.Button.LOG, isNew ? 0 : 1);
        buttons.put(FlowConst.Button.WORD, isNew || !MenuHelper.hasPermisson(mapPermisson, FlowConst.MenuFuncRight.EXP.value) ? 0 : 1);
        buttons.put(FlowConst.Button.EXCEL, isNew || !MenuHelper.hasPermisson(mapPermisson, FlowConst.MenuFuncRight.EXP.value) ? 0 : 1);
        buttons.put(FlowConst.Button.DEL, !isNew ? 1 : 0);
        buttons.put(FlowConst.Button.ADD, canAdd ? 1 : 0);
        //buttons.put(FlowConst.Button.DISUSE, canDis ? 1 : 0);
        buttons.put(FlowConst.Button.DISUSE, 0);

        //单据作废
        if (procInst.getStatu() == FlowConst.InstanceStatu.DISUSE.value) {
            buttons.put(FlowConst.Button.DISUSE, 0);
            buttons.put(FlowConst.Button.DEL, 0);
            context.getDfpResponse().put("buttons", buttons);
            return;
        }
        if (procInst.getStatu() == FlowConst.InstanceStatu.FINISH.value) {
            buttons.put(FlowConst.Button.DEL, 0);
        }
        ActivityEntity act = ActivityEntityBuffer.getInstance().get(actInst.getActId());
        //非制单状态时 编辑权限受 功能权限+流程修改权限+流程节点修改权限
        boolean notMakeEdit = act != null && act.getActEdit() && othCanEdit && MenuHelper.hasPermisson(mapPermisson, FlowConst.MenuFuncRight.UPD.value);
        boolean handler_make = isHandleMakeEx(act);

        //当前任务是我办理的(我的待办任务，前面读取的时候也已经写了办理人为我了)：
        if (act == null || act.getActType() == FlowConst.ActivityType.endEvent.value) {
            if (isAdmin) buttons.put(FlowConst.Button.RETAKE, 1);
            buttons.put(FlowConst.Button.DISUSE, 0);
        } else if (user_id.equalsIgnoreCase(actInst.getHandler()) || isAdmin) {
            if (actInst.getStatu() == FlowConst.ActivityStatu.WAIT.value) {
                //待办任务，显示办理按钮
                if (notMakeEdit || handler_make) {
                    buttons.put(FlowConst.Button.HANDLER, 1);
                    buttons.put(FlowConst.Button.SUBMIT, 0);
                } else {
                    buttons.put(FlowConst.Button.SUBMIT, 1);
                    buttons.put(FlowConst.Button.REJECT, 1);
                    buttons.put(FlowConst.Button.REJECT_MAKE, 1);
                }
            } else if (actInst.getStatu() == FlowConst.ActivityStatu.HANDLE.value) {
                buttons.put(FlowConst.Button.SUBMIT, 1);
                //办理中任务，提交/驳回
                if (actInst.isMake()) {
                    buttons.put(FlowConst.Button.SAVE, 1);
                    if (canDel && !isNew) {
                        buttons.put(FlowConst.Button.DEL, 1);
                    }
                } else {
                    if (notMakeEdit || handler_make) {
                        buttons.put(FlowConst.Button.SAVE, 1);
                    }
                    buttons.put(FlowConst.Button.REJECT, 1);
                    buttons.put(FlowConst.Button.REJECT_MAKE, 1);
                }
            } else if (actInst.getStatu() == FlowConst.ActivityStatu.SUBMIT.value) {
                //我已提交的任务，取回(前面已经过滤掉不能取回的任务：下一任务已办理)
                buttons.put(FlowConst.Button.RETAKE, 1);
            }
        } else if (actInst.getStatu() == FlowConst.ActivityStatu.WAIT.value && new ActinstDao().isMyWaitedTask(actInst.getTaskId(), user_id)) {
            //待办任务，显示办理按钮
            if (notMakeEdit || handler_make) {
                buttons.put(FlowConst.Button.HANDLER, 1);
                buttons.put(FlowConst.Button.SUBMIT, 0);
            } else buttons.put(FlowConst.Button.SUBMIT, 1);
        } else if (actInst.isMake() && actInst.getStatu() == FlowConst.ActivityStatu.HANDLE.value && canEdit && BillFlowWfHelper.getPropBool(procDef.getEntityId(), FlowConst.NULL_STR, FlowConst.ProcProperty.CAN_EDIT_OTHER)) {
            //不是我的任务，制单任务判断属性，是否可以修改他人单据
            ProcdefEntity def = BillFlowWfHelper.getBillProc(procInst.getBillType(), user_id);
            if (def != null && def.getEntityId().equals(procDef.getEntityId())) {
                buttons.put(FlowConst.Button.SAVE, 1);
                buttons.put(FlowConst.Button.SUBMIT, 1);
            }
        }
        context.getDfpResponse().put("buttons", buttons);
         */
    }

    /**
     * 校验权限
     *
     * @throws Exception
     */
    private void checkPermission() {
        if (actInst == null) {
            throw new SwException("没有找到对应的任务步骤！");
        }
        long user_id = us.getUserId();
        //是我办理的单据，正常，返回
        if (actInst.getHandler() <= 0L || user_id == actInst.getHandler()) return;
        //不是我办理的单据，非制单状态，不允许他人操作
        if (!actInst.isMake()) {
            throw new BizException("非制单步骤，必须办理人本人操作！");
        }
    }

    public void submitToEnd() {
        long user_id = us.getUserId();
        //校验权限，当前人员有权限干不，没权限直接抛异常
        checkPermission();
        if (actInst.getStatu() == FlowConst.ActivityStatu.SUBMIT.value) {
            throw new BizException("当前任务已经提交，不能重复提交");
        }

        if (actInst.getStatu() == FlowConst.ActivityStatu.WAIT.value && !FlowHelper.isMyWaitedTask(actInst.getEntityId(), user_id)) {
            throw new BizException("您不是当前任务的办理人，不允许提交");
        }
        //校验完毕，接下来干活。
        List<Activity> new_acts = procDef.getProcInfo().findEndActivity();
        if (CommUtil.isEmpty(new_acts)) {
            throw new BizException("结束任务为空，不允许提交到结束！");
        }

        //待新增任务
        List<Task> listInsertTask = new ArrayList<>();
        List<TaskRel> listTaskRel = new ArrayList<>();
        //待新增任务候选人
        List<Candidate> listInsertCand = new ArrayList<>();

        for (Activity act : new_acts) {
//            buildTask(act, listInsertTask, listInsertCand, listTaskRel, user_id);  todo
        }

        actInst.setHandler(user_id);
        actInst.setStartTime(DateUtil.nowDateTimeLong());
        actInst.setStatu(FlowConst.ActivityStatu.SUBMIT.value);
        actInst.setRetake(true);

        if (listInsertTask.size() > 0) {
            procInst.setTaskId(listInsertTask.get(0).getEntityId());
        } else {
            procInst.setTaskId(actInst.getEntityId());
        }

        if (new_acts.size() == 1 && new_acts.get(0).getType() == FlowConst.ActivityType.END.value) {
            procInst.setStatu(FlowConst.InstanceStatu.FINISH.value);
            procInst.setEndTime(DateUtil.nowDateTimeLong());
        } else {
            procInst.setStatu(FlowConst.InstanceStatu.RUNING.value);
        }

        //保存入库
        if (!listInsertTask.isEmpty()) {
            getTaskDao().batchInsertEntity(listInsertTask);
            getCandidateDao().batchInsertEntity(listInsertCand);
            getTaskRelDao().batchInsertEntity(listTaskRel);
        }
        getTaskDao().updateEntity(actInst);
        getProcInstDao().updateEntity(procInst);

        buildBillLog(FlowConst.Button.SUBMIT.value, actInst, listInsertTask);
    }


    /**
     * 提交
     *
     * @throws Exception
     */
    /*
    public void submit(AbsBillWfComponent comp) {
        String user_id = context.getLoginInfo().getUserId();
        //校验权限，当前人员有权限干不，没权限直接抛异常
        checkPermission();
        if (actInst.getStatu() == FlowConst.ActivityStatu.SUBMIT.value) {
            throw new DfpException("当前任务已经提交，不能重复提交");
        }

        ActinstDao actDao = new ActinstDao();
        if (actInst.getStatu() == FlowConst.ActivityStatu.WAIT.value && !actDao.isMyWaitedTask(actInst.getEntityId(), user_id) && !UserBaseEntityBuffer.getInstance().isAdmin(user_id)) {
            throw new DfpException("您不是当前任务的办理人，不允许提交");
        }
        //校验完毕，接下来干活。todo：流程定义需要控制一下，非判断/并发分支不允许多个出口
        List<ActivityEntity> new_acts = BillHelper.getNextActivity(comp);

        //待新增任务
        List<Task> listInsertTask = new ArrayList<>();
        List<ActinstRelEntity> listTaskRel = new ArrayList<>();
        //待新增任务候选人
        List<CandidateEntity> listInsertCand = new ArrayList<>();

        for (ActivityEntity act : new_acts) {
            buildTask(act, listInsertTask, listInsertCand, listTaskRel, user_id);
        }

        actInst.setHandler(user_id);
        actInst.setStartTime(UtilDateTime.nowDateTimeString());
        actInst.setstatu(FlowConst.ActivityStatu.SUBMIT.value);
        actInst.setRetake(true);

        if (listInsertTask.size() > 0) {
            procInst.setTaskId(listInsertTask.get(0).getEntityId());
            procInst.setTaskName(ActivityEntityBuffer.getInstance().getNameById(listInsertTask.get(0).getActId()));
        } else {
            procInst.setTaskId(actInst.getEntityId());
            procInst.setTaskName(ActivityEntityBuffer.getInstance().getNameById(actInst.getActId()));
        }

        if (new_acts.size() == 1 && new_acts.get(0).getActType() == FlowConst.ActivityType.endEvent.value) {
            procInst.setstatu(FlowConst.InstanceStatu.FINISH.value);
            procInst.setEndTime(UtilDateTime.nowDateTimeString());
        } else {
            procInst.setstatu(FlowConst.InstanceStatu.RUNING.value);
        }

        //保存入库
        if (!listInsertTask.isEmpty()) {
            actDao.batchInsert(listInsertTask);
            new CandidateDao().batchInsert(listInsertCand);
            new ActinstRelDao().batchInsert(listTaskRel);
        }
        actDao.update(actInst);
        new ProcinstDao().update(procInst);

        buildBillLog(FlowConst.FlowOptType.SUBMIT.value, actInst, listInsertTask);
        comp.transCallback(FlowConst.FlowOptType.SUBMIT.value, actInst, listInsertTask);
    }


    public void checkSubmit(AbsBillWfComponent comp) {
        String user_id = context.getLoginInfo().getUserId();
        //校验权限，当前人员有权限干不，没权限直接抛异常
        checkPermission();
        if (actInst.getStatu() == FlowConst.ActivityStatu.SUBMIT.value) {
            throw new DfpException("当前任务已经提交，不能重复提交");
        }
        ActinstDao actDao = new ActinstDao();
        if (actInst.getStatu() == FlowConst.ActivityStatu.WAIT.value && !actDao.isMyWaitedTask(actInst.getEntityId(), user_id) && !UserBaseEntityBuffer.getInstance().isAdmin(user_id)) {
            throw new DfpException("您不是当前任务的办理人，不允许提交");
        }
        comp.checkSubmit(FlowConst.FlowOptType.SUBMIT.value, actInst);
    }
    */

    /**
     * 作废
     *
     * @throws Exception
     */
    /*
    public void disuse(AbsBillWfComponent comp) {
        String user_id = context.getLoginInfo().getUserId();
        //校验权限，当前人员有权限干不，没权限直接抛异常
        checkPermission();
        if (procInst.getStatu() == FlowConst.InstanceStatu.FINISH.value) {
            throw new DfpException("当前单据已审批完成，不能作废单据");
        }

        ActinstDao actDao = new ActinstDao();
        if (actInst.getStatu() == FlowConst.ActivityStatu.WAIT.value && !actDao.isMyWaitedTask(actInst.getEntityId(), user_id) && !UserBaseEntityBuffer.getInstance().isAdmin(user_id)) {
            throw new DfpException("您不是当前任务的办理人，不允许作废");
        }

        actInst.setHandler(user_id);
        actInst.setStartTime(UtilDateTime.nowDateTimeString());
        actInst.setstatu(FlowConst.ActivityStatu.DISUSE.value);
        procInst.setstatu(FlowConst.InstanceStatu.DISUSE.value);
        procInst.setEndTime(UtilDateTime.nowDateTimeString());

        actDao.update(actInst);
        new ProcinstDao().update(procInst);

        buildBillLog(FlowConst.FlowOptType.DISUSE.value, actInst, null);
        comp.transCallback(FlowConst.FlowOptType.DISUSE.value, actInst, null);
    }

     */

    /**
     * 提交创建任务
     *
     * @param act
     * @param listInsertTask
     * @param listInsertCand
     * @param listTaskRel
     * @throws Exception
     */
    /*
    private void buildTask(ActivityEntity act, List<Task> listInsertTask, List<CandidateEntity> listInsertCand, List<ActinstRelEntity> listTaskRel, String submit_user) {
        Task task = new Task();
        listInsertTask.add(task);
        task.init();
        task.setEntityId(PKGenerator.newId());
        task.setBillId(procInst.getEntityId());
        task.setProcDefId(procDef.getEntityId());
        task.setActId(act.getEntityId());
        task.setStartTime(UtilDateTime.nowDateTimeString());
        task.setstatu(FlowConst.ActivityStatu.WAIT.value);
        task.setMake(false);
        task.setSign(false);
        task.setReject(false);
        task.setRetake(false);
        task.setAuto(false);

        ActinstRelEntity rel = new ActinstRelEntity();
        rel.setEntityId(PKGenerator.newId());
        rel.setBillId(task.getBillId());
        rel.setTaskId(task.getEntityId());
        rel.setBeforeTaskBid(actInst.getEntityId());
        listTaskRel.add(rel);

        if (act.getActType() != FlowConst.ActivityType.endEvent.value) {
            boolean handler_make = isHandleMakeEx(act);
            Set<String> listHander = new HashSet<>();
            if (handler_make) {
                listHander.add(procInst.getUserId());
            } else {
                String hanlder_group = BillFlowWfHelper.getPropStr(procInst.getProcDefId(), act.getEntityId(), FlowConst.ProcProperty.HANDLER);
                int handler_range = BillFlowWfHelper.getPropInt(procInst.getProcDefId(), act.getActId(), FlowConst.ProcProperty.HANDLER_RANGE);
                //获取所有候选人
                listHander = UserGroupHelper.getUserIdList(hanlder_group, procInst, handler_range, context);
            }
            if (UserBaseEntityBuffer.getInstance().isAdmin(context.getLoginInfo().getUserId()) && UtilPub.isEmpty(listHander)) {
                listHander = new HashSet<>();
                listHander.add(context.getLoginInfo().getUserId());
            }
            if (UtilPub.isEmpty(listHander)) {
                throw new DfpException("没有符合条件的主办人！" + act.getActName());
            }
            for (String user_id : listHander) {
                if (act.getActIgnore() && user_id.equals(submit_user)) continue;
                CandidateEntity cand = new CandidateEntity();
                cand.init();
                cand.setEntityId(PKGenerator.newId());
                cand.setBillId(task.getBillId());
                cand.setTaskId(task.getEntityId());
                cand.setUserId(user_id);

                listInsertCand.add(cand);
            }
            if (UtilPub.isEmpty(listInsertCand)) {
                throw new DfpException("没有符合条件的主办人！" + act.getActName());
            }
        }
    }

     */

    /**
     * 取回
     *
     * @throws Exception
     */
    /*
    public void retake(AbsBillWfComponent comp) {
        String user_id = context.getLoginInfo().getUserId();
        //校验权限，当前人员有权限干不，没权限直接抛异常
        checkPermission();
        if (actInst.getStatu() != FlowConst.ActivityStatu.SUBMIT.value && !UserBaseEntityBuffer.getInstance().isAdmin(user_id)) {
            throw new DfpException("当前任务非提交状态，不能取回！");
        }
        ActinstDao actDao = new ActinstDao();
        ActivityEntity activityEntity = ActivityEntityBuffer.getInstance().get(actInst.getActId());
        if (activityEntity.getActType() == FlowConst.ActivityType.endEvent.value) {//从结束取回
            List<Task> new_acts = actDao.findBeforeTasks(actInst.getEntityId());
            Task old_act = (Task) actInst.clone();
            old_act.setEntityId(actInst.getEntityId());

            actInst = new_acts.get(0);
            actInst.setHandler(user_id);
            actInst.setStartTime(UtilDateTime.nowDateTimeString());
            actInst.setstatu(FlowConst.ActivityStatu.HANDLE.value);
            if (actInst.isMake()) {
                actInst.setHandler(comp.billEntity.getBillUser());
            }

            procInst.setTaskId(actInst.getEntityId());
            procInst.setTaskName(ActivityEntityBuffer.getInstance().getNameById(actInst.getActId()));
            procInst.setstatu(actInst.isMake() ? FlowConst.InstanceStatu.BEGIN.value : FlowConst.InstanceStatu.RUNING.value);

            actDao.update(actInst);
            actDao.delete(old_act);
            new ActinstRelDao().deleteAll("where before_task_bid=?", old_act.getEntityId());
            new ProcinstDao().update(procInst);
            new_acts.clear();
            new_acts.add(old_act);
            buildBillLog(FlowConst.FlowOptType.RETAKE.value, actInst, new_acts);
            comp.transCallback(FlowConst.FlowOptType.RETAKE.value, actInst, new_acts);
        } else {

            List<Task> new_acts = actDao.findAfterTasks(actInst.getEntityId());

            for (Task act : new_acts) {
                if (act.getStatu() != FlowConst.ActivityStatu.WAIT.value) {
                    throw new DfpException("后续任务[" + ActivityEntityBuffer.getInstance().getNameById(act.getActId()) + "]已经办理，不能取回！");
                }
            }

            actInst.setHandler(user_id);
            actInst.setStartTime(UtilDateTime.nowDateTimeString());
            actInst.setstatu(FlowConst.ActivityStatu.HANDLE.value);

            procInst.setTaskId(actInst.getEntityId());
            procInst.setTaskName(ActivityEntityBuffer.getInstance().getNameById(actInst.getActId()));

            procInst.setstatu(FlowConst.InstanceStatu.RUNING.value);
            if (actInst.isMake()) procInst.setstatu(FlowConst.InstanceStatu.BEGIN.value);
            actDao.update(actInst);
            actDao.batchDelete(new_acts);
            new ActinstRelDao().deleteAll("where before_task_bid=?", actInst.getEntityId());
            new ProcinstDao().update(procInst);

            buildBillLog(FlowConst.FlowOptType.RETAKE.value, actInst, new_acts);
            comp.transCallback(FlowConst.FlowOptType.RETAKE.value, actInst, new_acts);
        }
    }

     */

    /**
     * 驳回
     *
     * @throws Exception
     */
    /*
    public void reject(AbsBillWfComponent comp) {
        String user_id = context.getLoginInfo().getUserId();
        //校验权限，当前人员有权限干不，没权限直接抛异常
        checkPermission();
        if (actInst.getStatu() == FlowConst.ActivityStatu.SUBMIT.value) {
            throw new DfpException("当前任务已提交，不能驳回！");
        }

        ActinstDao actDao = new ActinstDao();
        //获取前置任务，前置任务需要置为被驳回的办理状态，同时，前置任务的其他后置任务需要删除
        List<Task> new_acts = actDao.findBeforeTasks(actInst.getEntityId());
        if (UtilPub.isEmpty(new_acts)) {
            throw new DfpException("没有找到前续任务，驳回失败！");
        }
        //记录需要被删除后续任务的任务id
        StringBuilder sqlIn = new StringBuilder(128);
        //设为被驳回状态
        for (Task act : new_acts) {
            act.setstatu(FlowConst.ActivityStatu.HANDLE.value);
            act.setReject(true);

            sqlIn.append(",").append(act.getEntityId());
        }

        actInst.setstatu(FlowConst.ActivityStatu.REJECT.value);

        procInst.setTaskId(new_acts.get(0).getEntityId());
        procInst.setTaskName(ActivityEntityBuffer.getInstance().getNameById(new_acts.get(0).getActId()));
        if (new_acts.size() == 1 && new_acts.get(0).isMake()) {
            procInst.setstatu(FlowConst.InstanceStatu.BEGIN.value);
        } else {
            procInst.setstatu(FlowConst.InstanceStatu.RUNING.value);
        }
        //删除这些任务的后续任务
        List<Task> del_acts = actDao.findAfterTasks(sqlIn.substring(1));
        actDao.batchDelete(del_acts);
        actDao.batchUpdate(new_acts);
        new ActinstRelDao().deleteAll("where before_task_bid in (" + sqlIn.substring(1) + ")");
        new ProcinstDao().update(procInst);

        buildBillLog(FlowConst.FlowOptType.REJECT.value, actInst, new_acts);
        comp.transCallback(FlowConst.FlowOptType.REJECT.value, actInst, new_acts);
    }

     */

    /**
     * 驳回到制单
     *
     * @throws Exception
     */
    /*
    public void rejectToMake(AbsBillWfComponent comp) {
        //校验权限，当前人员有权限干不，没权限直接抛异常
        checkPermission();
        if (actInst.getStatu() == FlowConst.ActivityStatu.SUBMIT.value) {
            throw new DfpException("当前任务已提交，不能驳回！");
        }

        ActinstDao actDao = new ActinstDao();
        //获取前置任务，前置任务需要置为被驳回的办理状态，同时，前置任务的其他后置任务需要删除
        Task make_act = actDao.findMakeTask(actInst.getBillId());
        if (EntityUtil.isNull(make_act)) {
            throw new DfpException("没有找到制单任务，驳回到制单失败！");
        }

        //设为被驳回状态
        make_act.setstatu(FlowConst.ActivityStatu.HANDLE.value);
        make_act.setReject(true);

        actInst.setstatu(FlowConst.ActivityStatu.REJECT.value);

        procInst.setTaskId(make_act.getEntityId());
        procInst.setTaskName(ActivityEntityBuffer.getInstance().getNameById(make_act.getActId()));
        procInst.setstatu(FlowConst.InstanceStatu.BEGIN.value);

        actDao.update(make_act);
        //删除非制单步骤
        actDao.deleteAll("where bill_id = ? and task_id <> ? ", make_act.getBillId(), make_act.getEntityId());
        new ActinstRelDao().deleteAll("where bill_id = ? and task_id <> ?", make_act.getBillId(), make_act.getEntityId());
        new ProcinstDao().update(procInst);
        //删除候选人
        new CandidateDao().deleteAll(" where bill_id = ? and task_id <> ?", make_act.getBillId(), make_act.getEntityId());

        List<Task> make_acts = new ArrayList<>();
        make_acts.add(make_act);

        buildBillLog(FlowConst.FlowOptType.REJECT.value, actInst, make_acts);
        comp.transCallback(FlowConst.FlowOptType.REJECT.value, actInst, make_acts);
    }

     */

    /**
     * 办理
     *
     * @throws Exception
     */
    /*
    public void handler(AbsBillWfComponent comp) {
        String user_id = context.getLoginInfo().getUserId();
        //校验权限，当前人员有权限干不，没权限直接抛异常
        checkPermission();
        if (actInst.getStatu() != FlowConst.ActivityStatu.WAIT.value) {
            throw new DfpException("当前任务非待办状态，不能办理失败！");
        }

        ActinstDao actDao = new ActinstDao();

        actInst.setHandler(user_id);
        actInst.setStartTime(UtilDateTime.nowDateTimeString());
        actInst.setstatu(FlowConst.ActivityStatu.HANDLE.value);

        procInst.setTaskId(actInst.getEntityId());
        procInst.setTaskName(ActivityEntityBuffer.getInstance().getNameById(actInst.getActId()));

        procInst.setstatu(FlowConst.InstanceStatu.RUNING.value);

        actDao.update(actInst);
        new ProcinstDao().update(procInst);

        buildBillLog(FlowConst.FlowOptType.HANDLER.value, actInst, null);
//        comp.transCallback(FlowConst.FlowOptType.RETAKE.value, act_inst, null);
    }

     */

    /**
     * 状态变更，构建日志
     */
    protected void buildBillLog(final String flow_opt, Task srcTask, List<Task> dstTasks) {
        /*LogEntity logEntity = new LogEntity();
        logEntity.setSuggestion(context.getDfpRequest().getParams().getStrIgnoreNull("suggestion"));
        logEntity.setFlowOpt(flow_opt);

        logEntity.setBillId(proc_inst.getBillId());
        logEntity.setOptTime(UtilPub.getLastTime());

        UserBaseEntity user = UtilLogin.getLoginUser(context.getLoginInfo());
        logEntity.setUserId(user.getEntityId());
        logEntity.setUserName(user.getName());
        logEntity.setLogId(PKGenerator.newId());
        logEntity.setSignPicture(UserBaseEntityBuffer.getInstance().getUserCardEntity(user.getUserId()).getSignPicture());

        String src = null, dst = null;
        String step_name = null;
        StringBuilder sbDst = null;
        if (srcTask != null) {
            src = ActivityEntityBuffer.getInstance().getNameById(srcTask.getActId());
            step_name = ActivityEntityBuffer.getInstance().getStepName(srcTask.getActId());
        }
        if (UtilPub.isNotEmpty(dstTasks)) {
            sbDst = new StringBuilder();
            for (Task a : dstTasks) {
                sbDst.append("/").append(ActivityEntityBuffer.getInstance().getNameById(a.getActId()));
            }
            dst = sbDst.substring(1);
        }
        if (flow_opt == FlowConst.FlowOptType.RETAKE.value) {
            logEntity.setInfo(dst + "->" + src);
            logEntity.setStepName(step_name);
        } else if (dst != null) {
            logEntity.setInfo(src + "->" + dst);
            logEntity.setStepName(step_name);
        } else if (src != null) {
            logEntity.setInfo(src);
            logEntity.setStepName(step_name);
        } else {
            logEntity.setInfo("-");
        }
        new LogDao().insert(logEntity);
         */
    }

    private EntityDao<ProcInst> getProcInstDao() {
        return DbEngine.getInstance().findDao(ProcInst.class);
    }

    private EntityDao<Task> getTaskDao() {
        return DbEngine.getInstance().findDao(Task.class);
    }

    private EntityDao<Candidate> getCandidateDao() {
        return DbEngine.getInstance().findDao(Candidate.class);
    }

    private EntityDao<Care> getCareDao() {
        return DbEngine.getInstance().findDao(Care.class);
    }

    private EntityDao<Comment> getCommentDao() {
        return DbEngine.getInstance().findDao(Comment.class);
    }

    private EntityDao<Sign> getSignDao() {
        return DbEngine.getInstance().findDao(Sign.class);
    }

    private EntityDao<TaskRel> getTaskRelDao() {
        return DbEngine.getInstance().findDao(TaskRel.class);
    }
}
